# ---
# Copyright (C) 2011 Andreas Auras <yak54@inkennet.de>
# minor changes by carsten Presser c@rstenpresser.de
#
# This file is part of DFAtmo the driver for 'Atmolight' controllers for XBMC and xinelib based video players.
#
# DFAtmo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# DFAtmo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
#
#
# This _WAS_ a example for a Python script based DFAtmo output driver.
# The output driver writes data to the KarateLight device.
#
# The output driver scripts should be located within the DFAtmo XBMC addon
# installation in sub directory 'script.dfatmo/resources/lib/drivers'. 
#
# ---

import os, sys, time, array, serial

# Important: Don't change order of area list!
areas = (
    'top',
    'bottom',
    'left',
    'right',
    'center',
    'top_left',
    'top_right',
    'bottom_left',
    'bottom_right' )


logger = None

def log(logLevel, msg):
    logger.log(logLevel, 'KarateNative: %s' % msg)


class MyOutputDriver:
    def __init__(self, dfatmo):
        self.dfatmo = dfatmo
        self.fd = None
        self.areaChannels = None
        self.channels = 0
        self.nullData = None
        self.startTime = 0
        self.lastColors = None
        self.inputBuffer = bytearray(64)
        self.outputBuffer = bytearray(64)
        self.outputBuffer4 = bytearray(4)
        self.hardwareChannels = 0
        self.firmwareVersion = 0
        self.pwmValue = 0

    def driverError(self, msg):
        return self.dfatmo.driverError(msg)
                
    # return supported interface version
    def getInterfaceVersion(self):
        return 1

    # read answer of the KarateLight device, returns number of bytes read from the device
    def readBack(self):
        try:
            j = 0
            ret = 0
            while ret < 4:
                ret = self.fd.read(4)
            # TODO check i at least 4 bytes have been received..

            if ret < 4:
                log(logger.LOG_ERROR, "only received  %d bytes, expected at least 4 bytes" % ret)


            # copy anwer
            for i in ret:
                self.inputBuffer[j] = ord(i)
                j+=1
            
            
            if ((self.inputBuffer[1] != 0x01) & (self.inputBuffer[1] != 0x20)):
                log(logger.LOG_ERROR, "unexpected Returncode: %s" % hex(self.inputBuffer[1]))
                #for c in self.inputBuffer[0:5]:
                  #print hex(c)

            # more then 4 bytes received...
            if self.inputBuffer[3] > 0:
                ret = self.fd.read(self.inputBuffer[3])
                # copy answer..
                for i in ret:
                    self.inputBuffer[j] = ord(i)
                    j+=1

            return j
        except IOError as err:
            raise self.driverError('Error reading from %s fails: %s' % (self.fileName, err))

    # open device and configure for number of channels
    def openOutputDriver(self):
        
        self.fileName = self.dfatmo.getParm('driver_param')
        if self.fileName == '':
            self.fileName = '/dev/ttyACM0'
            
        try:
            # open the device
            self.fd = serial.Serial(self.fileName, timeout=1)

            # trash input chars
            r = 1
            while r:
              r = self.fd.read()

            # read Firmware Version
            self.outputBuffer[0] = 0xAA
            self.outputBuffer[1] = 0x01
            self.outputBuffer[2] = 0xAB
            self.outputBuffer[3] = 0x00
            self.fd.write(self.outputBuffer[0:4])
            self.fd.flush()
            if MyOutputDriver.readBack(self) == 5:
                self.firmwareVersion = self.inputBuffer[4]
                log(logger.LOG_INFO, "device reports firmware version %s " % hex(self.firmwareVersion))
            else:
                raise self.driverError('Failed to read firmware version')

            # read HardwareChannels 
            self.outputBuffer[0] = 0xAA
            self.outputBuffer[1] = 0x13
            self.outputBuffer[2] = 0xB9
            self.outputBuffer[3] = 0x00
            self.fd.write(self.outputBuffer[0:4])
            self.fd.flush()
            if MyOutputDriver.readBack(self) == 5:
                self.hardwareChannels = int(self.inputBuffer[4] * 8 / 12 / 3);
                log(logger.LOG_INFO, "device reports %d channels" % self.hardwareChannels)
            else:
                raise self.driverError('Failed to read number of channels')

        except IOError as err:
            raise self.driverError('Initialising device %s fails: %s' % (self.fileName, err))

        
        self.getAreasConfig()
        self.lastColors = None

        log(logger.LOG_INFO, "device '%s' opened" % self.fileName)
        self.startTime = time.time()
            
    # close device
    def closeOutputDriver(self):
        fd = self.fd
        if fd:
            self.fd = None
            try:
                fd.close()
            except IOError as err:
                raise self.driverError('Closing device %s fails: %s' % (self.fileName, err))

    def getAreasConfig(self):
        self.channels = 0
        self.areaChannels = dict()
        for area in areas:
            n = self.dfatmo.getParm(area)
            self.areaChannels[area] = n
            self.channels = self.channels + n

        if self.channels > self.hardwareChannels:
            raise self.driverError('Hardware has only %d channels, but  %d channels have been configured' % (self.hardwareChannels, self.channels))

        n = self.hardwareChannels * 3
        self.nullData = bytearray(n)
        while n:
            n = n - 1
            self.nullData[n] = 0


    # instant configure
    def instantConfigure(self):
        pass

    # turn all lights off                   
    def turnLightsOff(self):
        self.outputColors(self.nullData)
        
    # send RGB color values to device
    # order for 'colors' is: top 1,2,3..., bottom 1,2,3..., left 1,2,3..., right 1,2,3..., center, top left, top right, bottom left, bottom right
    # order of color packet: Red Green Blue
    def outputColors(self, colors):
        if self.lastColors != None and self.lastColors == colors:
            return
        self.lastColors = None

        i = 0
        for area in areas:
            areaChannels = self.areaChannels[area]
            areaNum = 1
            while areaChannels:
                self.outputBuffer[4+i+0] = colors[i+1] # Green first
                self.outputBuffer[4+i+1] = colors[i+2] # Blue
                self.outputBuffer[4+i+2] = colors[i+0] # Red
                areaChannels = areaChannels - 1
                areaNum = areaNum + 1
                i = i + 3

        # Padding with zero (black) if less then 8/16 channels are defined
        if i < (self.hardwareChannels*3):
            for j in range(i,self.hardwareChannels):
                self.outputBuffer[4+j] = 0;
            i = (self.hardwareChannels*3);

        # Prepare Prefix
        self.outputBuffer[0] = 0xAA
        self.outputBuffer[1] = 0x12
        self.outputBuffer[3] = i 

        # checksum
        self.outputBuffer[2] = self.outputBuffer[0] ^ self.outputBuffer[1]
        for j in range(3,i+4):
            self.outputBuffer[2] ^= self.outputBuffer[j]

        try:
            self.fd.write(self.outputBuffer[0:i+4])
            self.fd.flush()
            if MyOutputDriver.readBack(self) != 4:
                log(logger.LOG_INFO, "missed an ACK packet!")

        except IOError as err:
            raise self.driverError('Writing to devcie %s fails: %s' % (self.fileName, err))

        self.lastColors = colors


# return new output driver instance
def newOutputDriver(dfatmo, l):
    global logger
    logger = l
    return MyOutputDriver(dfatmo)
         
